module set_all_contours_m

  implicit none

contains

  subroutine set_all_contours(s, step, periodic, ssh, u, v, corner)

    ! This procedure sets all outermost contours in the snapshot
    ! "s".

    ! Libraries:
    use contour_531, only: convert_to_ind
    use jumble, only: argwhere, twopi, pi
    use numer_rec_95, only: indexx

    use config_m, only: min_amp, min_radius
    use derived_types, only: snapshot, null_ssh_contour, ssh_contour
    use get_1_outerm_m, only: get_1_outerm
    use input_ssh_m, only: max_radius
    use nearby_extr_m, only: nearby_extr
    use set_max_speed_m, only: set_max_speed

    type(snapshot), intent(inout):: s
    ! Define s%list%out_cont. s%number_extr, s%list%cyclone,
    ! s%list%extr, s%list%innermost_level, s%extr_map should be
    ! defined on entry.

    real, intent(in):: step(:) ! (2) longitude and latitude steps, in rad
    logical, intent(in):: periodic ! grid is periodic in longitude

    real, intent(in), dimension(merge(1 - max_radius(1), 1, periodic):, :):: &
         ssh, u, v
    ! (1 - max_radius(1):nlon + max_radius(1), nlat) if the grid is
    ! periodic in longitude, else (nlon, nlat). Sea-surface height, in
    ! m, and wind, in m s-1.

    real, intent(in):: corner(:) ! (2) longitude and latitude of the
    ! corner of the global grid, in rad

    ! Local:

    real min_area ! minimum area of an outermost contour, in m2
    integer nlon, nlat
    real innermost_level_2
    integer i, l
    integer n_cycl ! number of cyclones

    integer, allocatable:: sorted_extr(:) ! (s%number_extr)
    ! Sorted identifying number of extrema: first, minima sorted in
    ! order of decreasing SSH, and second, maxima sorted in order of
    ! increasing SSH.

    integer, allocatable:: selection(:)
    ! identifying numbers of a selection of extrema

    ! Window around each extremum:

    integer llc(2) ! indices in global grid of lower left corner
    integer urc(2) ! indices in global grid of upper right corner

    real corner_window(2) ! longitude and latitude of the window around each
    ! extremum, in rad

    real, allocatable:: outside_points(:, :) ! (2, :) longitude and
    ! latitude, in rad, of all the significant extrema, except the
    ! target extremum

    integer, parameter:: n_max_cont = 31

    type(ssh_contour) cont_list(n_max_cont)
    ! Contour list. The outermost contour is element with subscript
    ! n_cont. The contours are in monotonic order of ssh. If n_cont >=
    ! 2, cont_list(:n_cont - 1)%area is missing and cont_list(:n_cont
    ! - 1) is not oriented.

    integer n_cont ! number of good contours found and stored in cont_list, >= 0

    !--------------------------------------------------------------

    if (.not. periodic) nlon = size(ssh, 1)
    nlat = size(ssh, 2)
    allocate(sorted_extr(s%number_extr))
    min_area = pi * (min_radius * 1e3)**2

    forall (i = 1:s%number_extr)
       s%list(i)%out_cont%closed = .true.
       ! must be intialized to true because it is used in nearby_extr
    end forall

    ! Double sort of extrema on cyclonicity and SSH value:

    selection = argwhere(s%list%cyclone)
    n_cycl = size(selection)
    selection = selection(indexx(s%list(selection)%extr%ssh))
    sorted_extr(:n_cycl) = selection(n_cycl:1:- 1) ! descending order of ssh

    selection = argwhere(.not. s%list%cyclone)
    selection = selection(indexx(s%list(selection)%extr%ssh))
    sorted_extr(n_cycl + 1:) = selection

    ! Done sorting

    do l = 1, s%number_extr
       i = sorted_extr(l)

       associate (e => s%list(i))
         ! Define the geographical window around each eddy extremum:

         llc = e%extr%coord_proj - max_radius
         urc = e%extr%coord_proj + max_radius

         llc(2) = max(llc(2), 1)
         urc(2) = min(urc(2), nlat)

         if (.not. periodic) then
            llc(1) = max(llc(1), 1)
            urc(1) = min(urc(1), nlon)
         end if

         corner_window = corner + (llc - 1) * step

         ! Only look for good contours with amplitudes >= min_amp:
         innermost_level_2 = merge(e%extr%ssh + merge(min_amp, - min_amp, &
              e%cyclone), e%innermost_level, &
              abs(e%extr%ssh - e%innermost_level) < min_amp)

         outside_points = nearby_extr(s%extr_map(llc(1):urc(1), &
              llc(2):urc(2)), s%list, i)

         ! Shift the longitude of each point to a value close to the
         ! longitude of the target extremum:
         if (periodic) outside_points(1, :) = outside_points(1, :) &
              + ceiling((corner_window(1) - outside_points(1, :)) / twopi) &
              * twopi

         call get_1_outerm(e%out_cont, cont_list, n_cont, e%cyclone, &
              e%extr%coord, innermost_level_2, outside_points, &
              ssh(llc(1):urc(1), llc(2):urc(2)), corner_window, step, min_area)

         ! Done with outermost contour, now let us take care of the
         ! max-speed contour.
         if (e%out_cont%closed) then
            ! Restrict the field to the outermost contour:

            llc = floor(convert_to_ind(minval(e%out_cont%points, dim = 2), &
                 corner, step))

            urc = ceiling(convert_to_ind(maxval(e%out_cont%points, dim = 2), &
                 corner, step))

            ! Should have no effect except because of roundup error:
            urc(2) = min(urc(2), nlat)
            if (.not. periodic) urc(1) = min(urc(1), nlon)

            corner_window = corner + (llc - 1) * step

            call set_max_speed(e, e%extr%coord_proj - llc + 1, &
                 ssh(llc(1):urc(1), llc(2):urc(2)), &
                 u(llc(1):urc(1), llc(2):urc(2)), &
                 v(llc(1):urc(1), llc(2):urc(2)), corner_window, step, &
                 cont_list(:n_cont))
         end if
       end associate
    end do

  end subroutine set_all_contours

end module set_all_contours_m
