module set_max_speed_m

  implicit none

contains

  subroutine set_max_speed(e, ind_targ_extr, ssh, u, v, corner, step, cont_list)

    ! This procedure defines the components speed_cont, max_speed and
    ! radius4 of argument e. On return, e%speed_cont may be a null ssh
    ! contour but e%max_speed is never set to missing_speed. However,
    ! e%max_speed might be NaN if speed_outerm is NaN (with any value
    ! of e%radius4). The value of e%radius4 set here is always >=
    ! 1. On return, if e%speed_cont is not a null ssh contour then
    ! e%max_speed is the speed on e%speed_cont.

    ! The length of the interval of longitudes of the domain, step(1)
    ! * (size(ssh, 1) - 1), should be < 180 degrees, so that the
    ! geometrical processing done here is non-ambiguous. The
    ! longitudes of outside points should be in the interval
    ! [corner(1), corner(1) + step(1) * (size(ssh, 1) - 1)] as there
    ! will be no automatic shifting by a multiple of 360 degrees.

    use, intrinsic:: IEEE_ARITHMETIC, only: IEEE_IS_NAN

    ! Lbraries:
    use contour_531, only: convert_to_ind
    use geometry, only: polygon_point_dist_2d

    use ccw_orient_m, only: ccw_orient
    use derived_types, only: eddy, null_ssh_contour, missing_ssh, ssh_contour, &
         missing_speed
    use good_contour_m, only: good_contour
    use inside_4_m, only: inside_4
    use max_speed_contour_ssh_m, only: max_speed_contour_ssh
    use mean_speed_m, only: mean_speed
    use spher_polyline_area_m, only: spher_polyline_area

    type(eddy), intent(inout):: e
    ! e%out_cont, e%cyclone, e%extr%ssh and e%extr%coord
    ! should be defined on input.

    integer, intent(in):: ind_targ_extr(:) ! (2) indices of the target extremum

    real, intent(in):: ssh(:, :), u(:, :), v(:, :)
    ! The domain should be the bounding box of out_cont, because we do
    ! not exclude other extrema.

    real, intent(in):: corner(:) ! (2)
    ! longitude and latitude of the lower left corner of the grid, in rad

    real, intent(in):: step(:) ! (2) longitude and latitude steps, in rad

    type(ssh_contour), intent(in):: cont_list(:) ! (n_cont)
    ! Contour list. The outermost contour is element with subscript
    ! n_cont. The contours are in monotonic order of ssh. If n_cont >=
    ! 2, cont_list(:n_cont - 1)%area is missing and cont_list(:n_cont
    ! - 1) is not oriented. n_cont >= 1.

    ! Local:
    real speed(size(cont_list)) ! (n_cont) speed on the contour
    integer i, n_cont

    !---------------------------------------------------------------

    n_cont = size(cont_list)
    e%radius4 = - 1

    do i = 1, n_cont
       speed(i) = mean_speed(u, v, cont_list(i)%polyline, e%extr%coord, &
            corner, step)
    end do

    i = maxloc(abs(speed), dim = 1, mask = .not. IEEE_IS_NAN(speed))
    ! (The speed may be NaN when we are near land.)

    if (i == 0) then
       ! All speed values are NaN
       e%speed_cont = null_ssh_contour()
       e%max_speed = missing_speed
    else if (i == n_cont) then
       ! Maximum speed on the outermost contour
       e%speed_cont = null_ssh_contour()
       e%max_speed = speed(i)
    else
       e%speed_cont = cont_list(i)
       e%speed_cont%area = spher_polyline_area(e%speed_cont%polyline)
       call ccw_orient(e%speed_cont)
       e%max_speed = speed(i)
    end if

  end subroutine set_max_speed

end module set_max_speed_m
